class HFExcelStyle:
    __slots__ = [
        'workbook',
        'styles',
    ]

    def __init__(self, workbook):
        self.workbook = workbook
        self.styles = {}

    def add(self, name, style):
        self.styles[name] = self.workbook.add_format(style)
        return self

    def __getattr__(self, name):
        return self.styles.get(name)

    def set_defaults(self):
        grey_color = '#E0E0E0'
        orange_color = '#FFB266'

        decimal_format = '#,##0.00'
        no_decimal_format = '#,##0'

        self.add('headline', {
            'bold': 1,
            'font_size': 12,
            'font': 'Arial',
            'align': 'center'
        })

        self.add('bold', {
            'font': 'Arial',
            'bold': 1
        })

        self.add('center_bold', {
            'font': 'Arial',
            'bold': 1,
            'align': 'center'
        })

        self.add('grey', {
            'font': 'Arial',
            'bg_color': grey_color,
        })

        self.add('grey_bold', {
            'font': 'Arial',
            'bg_color': grey_color,
            'bold': 1
        })

        self.add('grey_bold_decimals', {
            'font': 'Arial',
            'bg_color': grey_color,
            'bold': 1,
            'num_format': decimal_format
        })

        self.add('grey_bold_no_decimals', {
            'font': 'Arial',
            'bg_color': grey_color,
            'bold': 1,
            'num_format': no_decimal_format
        })

        self.add('grey_center', {
            'font': 'Arial',
            'bg_color': grey_color,
            'align': 'center',
            'valign': 'vcenter',
        })

        self.add('grey_center_decimals', {
            'font': 'Arial',
            'bg_color': grey_color,
            'align': 'center',
            'num_format': decimal_format
        })

        self.add('grey_center_no_decimals', {
            'font': 'Arial',
            'bg_color': grey_color,
            'align': 'center',
            'num_format': no_decimal_format
        })

        self.add('grey_center_bold', {
            'font': 'Arial',
            'bg_color': grey_color,
            'align': 'center',
            'bold': 1,
            'valign': 'vcenter'
        })

        self.add('grey_center_percentage', {
            'font': 'Arial',
            'bg_color': grey_color,
            'align': 'center',
            'num_format': '0.00"%"'
        })

        self.add('no_decimal_format', {
            'font': 'Arial',
            'num_format': no_decimal_format
        })

        self.add('decimal_format', {
            'font': 'Arial',
            'num_format': decimal_format
        })

        self.add('arial', {
            'font': 'Arial'
        })

        self.add('align_right', {
            'font': 'Arial',
            'align': 'right'
        })

        self.add('align_left', {
            'font': 'Arial',
            'align': 'left'
        })

        self.add('align_right_bold', {
            'font': 'Arial',
            'align': 'right',
            'bold': 1
        })

        self.add('align_left_bold', {
            'font': 'Arial',
            'align': 'left',
            'bold': 1
        })

        self.add('align_date_left', {
            'font': 'Arial',
            'align': 'left',
            'num_format': self.date_format_excel
        })

        self.add('orange', {
            'bg_color': orange_color,
            'font_color': 'white',
            'bold': 1,
            'font': 'Arial',
            'align': 'center'
        })
        return self



from .models import HFExcelWorkbook


class HFExcel:
    __slots__ = []

    @staticmethod
    def hf_workbook(filename=None, set_default_styles=False):
        return HFExcelWorkbook(filename=filename,
                               set_default_styles=set_default_styles)



from jsonschema import validate

from xlsxwriter.utility import xl_rowcol_to_cell

from .schemas import DEFAULT_SCHEMA


def get_coor_name(from_row, from_col, to_row, to_col):
    start = xl_rowcol_to_cell(from_row, from_col)
    end = xl_rowcol_to_cell(to_row, to_col)
    return f"{start}:{end}"


class HFWorkbookFilter:
    __slots__ = ['bp_workbook']

    def __init__(self, bp_workbook):
        self.bp_workbook = bp_workbook

    def _raise_error(self, text):
        raise ValueError(text)

    def _populate_sheets_with_json(self, sheets):
        for sheet_json in sheets:
            sheet = self.bp_workbook.add_sheet(
                sheet_json.get('key'),
                name=sheet_json.get('name'),
                page_width=sheet_json.get('page_width'),
                page_height=sheet_json.get('page_height'))

            self._populate_columns_with_json(sheet,
                                             sheet_json.get('columns',
                                                            []))

    def _populate_columns_with_json(self, sheet, columns):
        for column_json in columns:
            width = int(column_json.get('width', 0)) or None
            column_name = column_json.get('name') or self._raise_error(
                'column name is missing')
            cell_format = column_json.get('cell_format')
            options = column_json.get('options')
            args = column_json.get('args', ())
            column, _ = sheet.add_column(*args,
                                         name=column_name,
                                         width=width,
                                         cell_format=cell_format,
                                         options=options)

            self._populate_rows_with_json(column,
                                          column_json.get('rows', []))

    def _populate_rows_with_json(self, column, rows):
        for row_json in rows:
            data = row_json.get('data') or self._raise_error(
                'row data is missing')
            args = row_json.get('args', ())
            width = int(row_json.get('width', 0)) or None
            row, _ = column.add_row(*args, data=data, width=width)

    def _populate_styles_with_json(self, styles):
        for style_json in styles:
            self.bp_workbook.add_style(style_json.get('name'),
                                       style_json.get('style'))

    def populate_with_json(self, workbook_data, schema=None):
        validate(instance=workbook_data, schema=(schema or DEFAULT_SCHEMA))
        self._populate_sheets_with_json(workbook_data.get('sheets', []))
        self._populate_styles_with_json(workbook_data.get('styles', []))
        return self



DEFAULT_SCHEMA = {
    "type": "object",
    "properties": {
        "sheets": {
            "type": "array",
            "items": {
                "allOf": [
                    {
                        "type": "object",
                        "properties": {
                            "key": {"type": "string"},
                            "name": {"type": "string"},
                            "args": {
                                "type": "array",
                                "items": {
                                    "oneOf": [
                                        {"type": "string"},
                                        {"type": "integer"}
                                    ]
                                }
                            },
                            "columns": {
                                "type": "array",
                                "items": {
                                    "allOf": [
                                        {
                                            "type": "object",
                                            "properties": {
                                                "name": {"type": "string"},
                                                "width": {"type": "integer"},
                                                "args": {
                                                    "type": "array",
                                                    "items": {
                                                        "oneOf": [
                                                            {"type": "string"},
                                                            {"type": "integer"}
                                                        ]
                                                    }
                                                },
                                                "rows": {
                                                    "type": "array",
                                                    "items": {
                                                        "allOf": [
                                                            {
                                                                "type": "object",
                                                                "properties": {
                                                                    "width": {
                                                                        "type": "integer"},
                                                                    "data": {
                                                                        "type": "string"},
                                                                    "args": {
                                                                        "type": "array"}
                                                                },
                                                                "required": [
                                                                    'data']
                                                            }
                                                        ]
                                                    }
                                                },
                                            },
                                            "required": ["rows"]
                                        }
                                    ]
                                }
                            }
                        }
                    }
                ]
            },
            "required": ["key", "columns"]
        },
        "styles": {
            "type": "array",
            "items": {
                "allOf": [{
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "style": {
                            "type": "object",
                            "additionalProperties": True
                        }
                    },
                    "required": ["name", "style"]
                }]
            }
        }
    },
    "required": ["sheets"]
}



from io import BytesIO

import xlsxwriter

from .helpers import get_coor_name, HFWorkbookFilter
from .styles import HFExcelStyle


class HFExcelWorkbook:
    __slots__ = [
        'filename',
        'output',
        '_sheets',
        '_set_default_styles',
        '_style',
        '_workbook',
    ]

    def __init__(self, filename=None, set_default_styles=True):
        self._sheets = []
        self._workbook = None
        self._style = None
        self.filename = filename
        self._set_default_styles = set_default_styles
        self._init_workbook(filename, set_default_styles)


    def filter(self):
        return HFWorkbookFilter(self)

    def _init_workbook(self, filename, set_default_styles):
        if filename and isinstance(filename, str):
            _workbook = xlsxwriter.Workbook(filename)
            self.output = None
        else:
            self.output = BytesIO()
            _workbook = xlsxwriter.Workbook(self.output)

        del self._workbook
        del self._style

        self._workbook = _workbook

        self._style = HFExcelStyle(self.workbook)
        if set_default_styles:
            self.set_default_styles()

    @property
    def workbook(self):
        return self._workbook

    @property
    def style(self):
        return self._style

    @property
    def hf_sheets(self):
        return self._sheets

    def add_style(self, name, style):
        self.style.add(name, style)

    def set_default_styles(self):
        return self.style.set_defaults()

    def add_sheet(self, key, name=None, columns=None, page_width=None,
                  page_height=None):
        for s in self._sheets:
            if s.key == key:
                return None
        sheet = HFExcelSheet(self,
                             key=key,
                             name=name,
                             columns=columns,
                             page_width=page_width,
                             page_height=page_height)
        self.hf_sheets.append(sheet)
        return sheet

    def hf_sheet(self, key):
        for s in self.hf_sheets:
            if s.key == key:
                return s
        return None

    def clean(self, filename=None, set_default_styles=None):
        self._sheets = []
        new_filename = filename or self.filename
        new_default_styles = (set_default_styles
                              or self._set_default_styles)
        self._init_workbook(new_filename, new_default_styles)

    def _save_sheets(self):
        next_col_index = 0
        for hf_sheet in self.hf_sheets:
            hf_sheet.sheet.print_area(0, 0,
                                      hf_sheet.height,
                                      hf_sheet.width)
            hf_sheet.sheet.fit_to_pages(hf_sheet.page_width,
                                        hf_sheet.page_height)
            next_col_index = hf_sheet.save(next_col_index)
        return True

    def save(self, close=True):
        self._save_sheets()
        if close:
            self.workbook.close()
        return self.workbook


class HFExcelSheet:
    __slots__ = [
        'key',
        'name',
        'page_height',
        'page_width',
        '_workbook',
        '_sheet',
        '_columns',
        '_curr_column_pos',
        '_curr_row_height',
    ]

    def __init__(self, workbook, key, name=None, columns=None, page_width=None,
                 page_height=None):
        self._workbook = workbook
        self.key = key
        self.name = name if name else key
        self._columns = columns or []
        self._curr_column_pos = 0
        self._curr_row_height = 0
        self._sheet = self.workbook.add_worksheet()
        self.page_width = int(page_width or 1)
        self.page_height = int(page_height or 0)

    def __getitem__(self, key):
        return self.columns[key]

    @property
    def workbook(self):
        return self.hf_workbook.workbook

    @property
    def hf_workbook(self):
        return self._workbook

    @property
    def sheet(self):
        return self._sheet

    @property
    def width(self):
        return self._curr_column_pos

    @property
    def height(self):
        return self._curr_row_height

    @property
    def row_height(self):
        return self._curr_row_height

    @property
    def next_row(self):
        return self.row_height + 1

    @property
    def next_col(self):
        return self._curr_column_pos

    @property
    def columns(self):
        return self._columns

    def add_column(self,
                   *args,
                   name='',
                   width=None,
                   cell_format=None,
                   options=None,
                   hide_header=False):

        width = width or 1
        if width <= 0:
            raise ValueError('Column width cannot be equal or less than 0')

        self._curr_column_pos = self._curr_column_pos + width

        column = HFExcelColumn(self,
                               *args,
                               name=name,
                               width=width,
                               cell_format=cell_format,
                               options=options,
                               hide_header=hide_header)

        self._columns.append(column)
        return column, len(self.columns) - 1

    def add_row(self, column_index, *args, data='', width=None):
        column = self.columns[column_index]
        row = column.add_row(*args, data=data, width=None)
        self._curr_row_height = max(
            column.hf_sheet.next_row - 1,
            len(column.rows)
        )
        return row

    def save(self, reference_index):
        next_col_index = reference_index
        next_row_index_max = 0
        for column in self.columns:
            next_col_index, next_row_index = column.save(next_col_index)
            next_row_index_max = max(next_row_index, next_row_index_max)
        return next_col_index, next_row_index_max


class HFExcelColumn:
    __slots__ = [
        'args',
        'cell_format',
        'hide_header',
        'name',
        'options',
        '_style',
        '_rows',
        '_sheet',
        '_width',
    ]

    def __init__(self,
                 sheet,
                 *args,
                 name='',
                 width=None,
                 cell_format=None,
                 options=None,
                 hide_header=False):
        self._sheet = sheet
        self._width = width
        self._style = (len(args) and args[0]) or None
        self.args = args and args[1:]
        self.cell_format = cell_format
        self.name = name
        self.options = options
        self.hide_header = hide_header
        self._rows = []

    def __getitem__(self, key):
        return self.rows[key]

    @property
    def required_args(self):
        return (self.name, self.style)

    @property
    def style(self):
        if not self._style:
            return None
        return getattr(self.hf_workbook.style, self._style)


    @property
    def width(self):
        return int(self._width or 1)

    @property
    def hf_sheet(self):
        return self._sheet

    @property
    def hf_workbook(self):
        return self.hf_sheet.hf_workbook

    @property
    def rows(self):
        return self._rows

    def add_row(self, *args, data='', width=None):
        row = HFExcelRow(self, *args, data=data, width=width)
        self.rows.append(row)
        return row, len(self.rows) - 1

    def save(self, reference_index):
        if self.hide_header:
            next_row = 0
            next_row_max = 0
        else:
            self.hf_sheet.sheet.set_column(reference_index,
                                           reference_index + self.width - 1)
            if self.width <= 1:
                new_args = self.required_args + self.args
                self.hf_sheet.sheet.write(0,
                                          reference_index,
                                          *new_args)
            else:
                coor_name = get_coor_name(0, reference_index,
                                          0, reference_index + self.width - 1)
                self.hf_sheet.sheet.merge_range(coor_name, self.name, self.style)
            next_row = 1
            next_row_max = 1
        for row in self.rows:
            _, next_row = row.save(reference_index, next_row)
            next_row_max = max(next_row, next_row_max)
        return reference_index + self.width, next_row_max


class HFExcelRow:
    __slots__ = [
        'args',
        'data',
        '_style',
        '_column',
        '_width',
    ]

    def __init__(self, column, *args, data='', width=None):
        self._column = column
        self.data = data
        self._style = (len(args) and args[0]) or None
        self.args = args and args[1:]
        self._width = width or 1

    @property
    def column(self):
        return self._column

    @property
    def width(self):
        if self._width > self.column.width:
            return self.column.width
        return self._width

    @property
    def hf_sheet(self):
        return self.column.hf_sheet

    @property
    def hf_workbook(self):
        return self.hf_sheet.hf_workbook

    @property
    def style(self):
        if not self._style:
            return None
        return getattr(self.hf_workbook.style, self._style)

    @property
    def required_args(self):
        return (self.data, self.style)

    def save(self, column, row):
        if self.width <= 1:
            new_args = self.required_args + self.args
            self.hf_sheet.sheet.write(row,
                                      column,
                                      *new_args)
        else:
            coor_name = get_coor_name(row, column,
                                      row, column + self.width - 1)
            self.hf_sheet.sheet.merge_range(coor_name, self.data, self.style)
        return column + self.width, row + 1



from .helpers import HFWorkbookFilter
from .managers import HFExcel
from .models import (HFExcelWorkbook,
                    HFExcelSheet,
                    HFExcelRow,
                    HFExcelColumn,
                    HFExcelStyle)



from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='hfexcel',
      version='0.0.12',
      description='human friendly excel creation in python',
      long_description=readme(),
      long_description_content_type='text/markdown',
      classifiers=[
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Text Processing :: Linguistic',
         'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      keywords='xlsxwriter xlsx excel json writing python wrapper',
      url='https://github.com/ebsaral/hfexcel',
      author='Emin Bugra Saral',
      author_email='eminbugrasaral@me.com',
      license='BSD',
      packages=['hfexcel'],
      install_requires=[
          'xlsxwriter',
          'jsonschema==2.6.0',
          'pytest',
      ],
      include_package_data=True,
      zip_safe=False,
      project_urls={
        'Documentation': 'https://github.com/ebsaral/hfexcel',
        'Funding': 'https://github.com/ebsaral/hfexcel',
        'Source': 'https://github.com/ebsaral/hfexcel',
      },
      tests_require=["pytest"],
)